@@ -2,6 +2,7 @@ import torch
 from torch import nn, einsum
 from inspect import isfunction
 from functools import partial
+from itertools import islice, cycle
 import torch.nn.functional as F
 
 from math import sqrt
@@ -9,7 +10,7 @@ from einops import rearrange, repeat, reduce
 from einops.layers.torch import Rearrange
 
 import alphafold2_pytorch.constants as constants
-from alphafold2_pytorch.utils import get_bucketed_distance_matrix, center_distogram_torch, MDScaling, scn_backbone_mask, scn_cloud_mask, sidechain_container
+from alphafold2_pytorch.utils import *
 from alphafold2_pytorch.reversible import ReversibleSequence
 
 # structure module
@@ -55,6 +56,92 @@ class PreNormCross(nn.Module):
         context = self.norm_context(context)
         return self.fn(x, context, *args, **kwargs)
 
+class InterceptAxialAttention(nn.Module):
+    def __init__(
+        self,
+        slice_tuple = None,
+        attn = None
+    ):
+        super().__init__()
+        assert exists(attn), 'attention function not given'
+        self.attn = attn
+        self.slice_tuple = slice_tuple
+
+    def forward(self, x, *args, shape = None, mask = None, **kwargs):
+        assert exists(shape), 'sequence shape must be given for intercepting and slicing of inputs'
+        attn, slice_tuple = self.attn, self.slice_tuple
+
+        if not exists(slice_tuple):
+            return self.attn(x, *args, shape = shape, mask = mask, **kwargs)
+
+        x = x.view(shape)
+        output = torch.zeros_like(x)
+        x = x[slice_tuple]
+        output_subset_shape = x.shape
+        x = rearrange(x, 'b ... d -> b (...) d')
+
+        if exists(mask):
+            mask = mask.view(shape[:-1])
+            mask = mask[slice_tuple]
+            mask = rearrange(mask, 'b ... -> b (...)')
+
+        attn_output = self.attn(x, *args, shape = output_subset_shape, mask = mask, **kwargs)
+
+        attn_output = attn_output.view(output_subset_shape)
+        output[slice_tuple] = attn_output
+        return rearrange(output, 'b ... d -> b (...) d')
+
+class InterceptAttention(nn.Module):
+    def __init__(
+        self,
+        slice_tuple,
+        attn = None,
+        context = False,
+    ):
+        super().__init__()
+        assert exists(attn), 'attention function not given'
+        self.attn = attn
+        self.slice_tuple = slice_tuple
+        self.context = context
+
+    def forward(self, *args, shape = None, context_shape = None, mask = None, context_mask = None, **kwargs):
+        assert exists(shape) and exists(context_shape), 'sequence and context shape must be given for intercepting and slicing of inputs'
+        attn, slice_tuple, context = self.attn, self.slice_tuple, self.context
+
+        x, c, *args = args
+
+        if context:
+            c = c.view(context_shape)
+            c = c[slice_tuple]
+            c = rearrange(c, 'b ... d -> b (...) d')
+
+            if exists(context_mask):
+                context_mask = context_mask.view(context_shape[:-1])
+                context_mask = context_mask[slice_tuple]
+                context_mask = rearrange(context_mask, 'b ... -> b (...)')
+        else:
+            x = x.view(shape)
+            output = torch.zeros_like(x)
+            x = x[slice_tuple]
+            output_subset_shape = x.shape
+            x = rearrange(x, 'b ... d -> b (...) d')
+
+            if exists(mask):
+                mask = mask.view(shape[:-1])
+                mask = mask[slice_tuple]
+                mask = rearrange(mask, 'b ... -> b (...)')
+
+        attn_output = self.attn(x, c, *args, mask = mask, context_mask = context_mask, **kwargs)
+
+        if context:
+            return attn_output
+
+        attn_output = attn_output.view(output_subset_shape)
+        output[slice_tuple] = attn_output
+        return rearrange(output, 'b ... d -> b (...) d')
+
+# feed forward
+
 class GEGLU(nn.Module):
     def forward(self, x):
         x, gates = x.chunk(2, dim = -1)
@@ -108,7 +195,7 @@ class Attention(nn.Module):
 
         self.tie_attn_dim = tie_attn_dim
 
-    def forward(self, x, context = None, mask = None, context_mask = None, tie_attn_dim = None):
+    def forward(self, x, context = None, mask = None, context_mask = None, tie_attn_dim = None, **kwargs):
         device, orig_shape, h, has_context = x.device, x.shape, self.heads, exists(context)
 
         context = default(context, x)
@@ -213,7 +300,7 @@ class SparseAttention(Attention):
             attn_mask_mode = 'add'
         )
 
-    def forward(self, x, mask = None):
+    def forward(self, x, mask = None, **kwargs):
         device, orig_shape, h = x.device, x.shape, self.heads
 
         b, n, _ = x.shape
@@ -246,6 +333,7 @@ class SparseAttention(Attention):
 class AxialAttention(nn.Module):
     def __init__(
         self,
+        template_axial_attn = False,
         tie_row_attn = False,
         sparse_attn = False,
         **kwargs
@@ -258,38 +346,91 @@ class AxialAttention(nn.Module):
         self.attn_width = attn_class(**kwargs)
         self.attn_height = attn_class(**kwargs)
 
-    def forward(self, x, shape, context = None, mask = None, context_mask = None):
+        self.template_axial_attn = template_axial_attn
+        if template_axial_attn:
+            self.attn_frames = Attention(**kwargs)
+
+    def forward(self, x, shape, mask = None):
         b, n, d = x.shape
 
+        num_axials = len(shape) - 2
         x = x.view(shape)
-        h, w = shape[1:3]
 
-        mask = mask.reshape(b, h, w) if exists(mask) else None
+        h, w = shape[-3:-1]
 
-        w_mask = h_mask = w_context = h_context = w_context_mask = h_context_mask = None
+        if num_axials == 3:
+            t = shape[1]
+        else:
+            t = 1
+            x = rearrange(x, 'b h w d -> b () h w d')
+            mask = rearrange(mask, 'b n -> b () n')
+
+        w_mask = h_mask = None
 
         if exists(mask):
-            w_mask = rearrange(mask, 'b h w -> (b w) h', w = w)
-            h_mask = rearrange(mask, 'b h w -> (b h) w', h = h)
+            mask = mask.reshape(b, t, h, w)
+            w_mask = rearrange(mask, 'b t h w -> (b t w) h')
+            h_mask = rearrange(mask, 'b t h w -> (b t h) w')
 
-        if exists(context):
-            w_context = repeat(context, 'b n d -> (b w) n d', w = w)
-            h_context = repeat(context, 'b n d -> (b h) n d', h = h)
-            w_context_mask = repeat(context_mask, 'b n -> (b w) n', w = w)
-            h_context_mask = repeat(context_mask, 'b n -> (b h) n', h = h)
+            if num_axials == 3:
+                f_mask = rearrange(mask, 'b t h w -> (b h w) t')
 
-        attn_kwargs = {} if not exists(context) else {'context': w_context, 'context_mask': w_context_mask}
-        w_x = rearrange(x, 'b h w d -> (b w) h d')
-        w_out = self.attn_width(w_x, mask = w_mask, **attn_kwargs)
-        w_out = rearrange(w_out, '(b w) h d -> b h w d', h = h, w = w)
+        w_x = rearrange(x, 'b t h w d -> (b t w) h d')
+        w_out = self.attn_width(w_x, mask = w_mask)
+        w_out = rearrange(w_out, '(b t w) h d -> b t h w d', h = h, w = w, t = t)
 
-        tie_attn_dim = x.shape[1] if self.tie_row_attn else None
-        h_x = rearrange(x, 'b h w d -> (b h) w d')
-        h_out = self.attn_height(h_x, mask = h_mask, tie_attn_dim = tie_attn_dim, **attn_kwargs)
-        h_out = rearrange(h_out, '(b h) w d -> b h w d', h = h, w = w)
+        tie_attn_dim = x.shape[2] if self.tie_row_attn else None
+        h_x = rearrange(x, 'b t h w d -> (b t h) w d')
+        h_out = self.attn_height(h_x, mask = h_mask, tie_attn_dim = tie_attn_dim)
+        h_out = rearrange(h_out, '(b t h) w d -> b t h w d', h = h, w = w, t = t)
 
         out = w_out + h_out
-        return rearrange(out, 'b h w d -> b (h w) d')
+
+        if self.template_axial_attn:
+            f_x = rearrange(x, 'b t h w d -> (b h w) t d')
+            f_out = self.attn_frames(f_x, mask = f_mask)
+            f_out = rearrange(f_out, '(b h w) t d -> b t h w d', h = h, w = w, t = t)
+
+            out += f_out
+
+        denom = 3 if self.template_axial_attn else 2
+        out /= denom
+
+        return rearrange(out, 'b t h w d -> b (t h w) d')
+
+# template module helpers and classes
+
+class SE3TemplateEmbedder(nn.Module):
+    def __init__(self, *args, **kwargs):
+        super().__init__()
+        self.net = SE3Transformer(*args, **kwargs)
+        self.sidechains_proj = nn.Parameter(torch.randn(1, kwargs['dim']))
+
+    def forward(
+        self,
+        t_seq,
+        templates_sidechains,
+        templates_coors,
+        mask = None
+    ):
+        shape = t_seq.shape
+        t_seq = rearrange(t_seq, 'b t n d-> (b t) n d ()')
+
+        templates_sidechains = rearrange(templates_sidechains, 'b t n m -> (b t) n () m')
+        templates_sidechains = einsum('b n d m, d e -> b n e m', templates_sidechains, self.sidechains_proj)
+        templates_coors = rearrange(templates_coors, 'b t n m -> (b t) n m')
+
+        mask = rearrange(mask, 'b t n -> (b t) n')
+
+        t_seq = self.net(
+            {'0': t_seq, '1': templates_sidechains},
+            templates_coors,
+            mask = mask,
+            return_type = 0
+        )
+
+        t_seq = t_seq.reshape(*shape)
+        return t_seq
 
 # structure module helpers and classes
 
@@ -332,15 +473,15 @@ class SequentialSequence(nn.Module):
 
             # self attention
 
-            x = attn(x, seq_shape, mask = mask) + x
+            x = attn(x, shape = seq_shape, mask = mask) + x
 
             if exists(m):
-                m = msa_attn(m, msa_shape, mask = msa_mask) + m
+                m = msa_attn(m, shape = msa_shape, mask = msa_mask) + m
 
                 # cross attention
 
-                x = cross_attn(x, m, mask = mask, context_mask = msa_mask) + x
-                m = msa_cross_attn(m, x, mask = msa_mask, context_mask = mask) + m
+                x = cross_attn(x, m, mask = mask, context_mask = msa_mask, shape = seq_shape, context_shape = msa_shape) + x
+                m = msa_cross_attn(m, x, mask = msa_mask, context_mask = mask, shape = msa_shape, context_shape = seq_shape) + m
 
             # feedforwards
 
@@ -360,6 +501,7 @@ class Alphafold2(nn.Module):
         depth = 6,
         heads = 8,
         dim_head = 64,
+        attn_types = ('full',),
         num_tokens = constants.NUM_AMINO_ACIDS,
         num_embedds = constants.NUM_EMBEDDS_TR,
         max_num_msas = constants.MAX_NUM_MSA,
@@ -405,19 +547,28 @@ class Alphafold2(nn.Module):
 
         # template sidechain encoding
 
-        self.sidechains_proj = nn.Parameter(torch.randn(1, dim))
+        self.use_se3_transformer = use_se3_transformer
 
-        self.template_sidechain_emb = SE3Transformer(
-            dim = dim,
-            dim_head = dim,
-            heads = 1,
-            num_neighbors = 12,
-            depth = 4,
-            input_degrees = 2,
-            num_degrees = 2,
-            output_degrees = 1,
-            reversible = True
-        )
+        if use_se3_transformer:
+            self.template_sidechain_emb = SE3TemplateEmbedder(
+                dim = dim,
+                dim_head = dim,
+                heads = 1,
+                num_neighbors = 12,
+                depth = 4,
+                input_degrees = 2,
+                num_degrees = 2,
+                output_degrees = 1,
+                reversible = True
+            )
+        else:
+            self.template_sidechain_emb = EnTransformer(
+                dim = dim,
+                dim_head = dim,
+                heads = 1,
+                num_nearest_neighbors = 32,
+                depth = 4
+            )
 
         # custom embedding projection
 
@@ -428,24 +579,27 @@ class Alphafold2(nn.Module):
         prenorm = partial(PreNorm, dim)
         prenorm_cross = partial(PreNormCross, dim)
 
-        template_layers = nn.ModuleList([])
-        for _ in range(template_attn_depth):
-            template_layers.append(nn.ModuleList([
-                prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
-                prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
-                prenorm(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout)),
-                prenorm(FeedForward(dim = dim, dropout = ff_dropout))
-            ]))
-
-        self.template_attn_net = template_layers
-
         layers = nn.ModuleList([])
-        for _, layer_sparse_attn in zip(range(depth), layers_sparse_attn):
+        attn_types = islice(cycle(attn_types), depth)
 
-            # self attention, for both main sequence and msa
+        for _, layer_sparse_attn, attn_type in zip(range(depth), layers_sparse_attn, attn_types):
+
+            # self attention, for main sequence, msa, and optionally, templates
+
+            if attn_type == 'full':
+                tensor_slice = None
+                template_axial_attn = True
+            elif attn_type == 'intra_attn':
+                tensor_slice = None
+                template_axial_attn = False
+            elif attn_type == 'seq_only':
+                tensor_slice = (slice(None), slice(0, 1))
+                template_axial_attn = False
+            else:
+                raise ValueError(f'cannot find attention type {attn_type}')
 
             layers.append(nn.ModuleList([
-                prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, sparse_attn = sparse_self_attn)),
+                prenorm(InterceptAxialAttention(tensor_slice, AxialAttention(dim = dim, template_axial_attn = template_axial_attn, seq_len = (max_seq_len, max_seq_len, max_num_templates), heads = heads, dim_head = dim_head, dropout = attn_dropout, sparse_attn = sparse_self_attn))),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
                 prenorm(AxialAttention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, tie_row_attn = msa_tie_row_attn)),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
@@ -453,10 +607,12 @@ class Alphafold2(nn.Module):
 
             # cross attention, for main sequence -> msa and then msa -> sequence
 
+            intercept_fn = partial(InterceptAttention, (slice(None), slice(0, 1)))
+
             layers.append(nn.ModuleList([
-                prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio)),
+                intercept_fn(context = False, attn = prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio))),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
-                prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio)),
+                intercept_fn(context = True, attn = prenorm_cross(Attention(dim = dim, seq_len = max_seq_len, heads = heads, dim_head = dim_head, dropout = attn_dropout, compress_ratio = cross_attn_compress_ratio))),
                 prenorm(FeedForward(dim = dim, dropout = ff_dropout)),
             ]))
 
@@ -469,23 +625,18 @@ class Alphafold2(nn.Module):
         # to distogram output
 
         self.num_backbone_atoms = num_backbone_atoms
+        needs_upsample = num_backbone_atoms > 1
 
-        if num_backbone_atoms > 1:
-            self.backbone_pos_emb = nn.Parameter(torch.randn(num_backbone_atoms, structure_module_dim))
-
-            self.to_distogram_logits = nn.Sequential(
-                nn.LayerNorm(dim),
+        self.to_distogram_logits = nn.Sequential(
+            nn.LayerNorm(dim),
+            nn.Sequential(
                 nn.Linear(dim, dim * (num_backbone_atoms ** 2)),
                 Rearrange('b h w c -> b c h w'),
                 nn.PixelShuffle(num_backbone_atoms),
-                Rearrange('b c h w -> b h w c'),
-                nn.Linear(dim, constants.DISTOGRAM_BUCKETS)
-            )
-        else:
-            self.to_distogram_logits = nn.Sequential(
-                nn.LayerNorm(dim),
-                nn.Linear(dim, constants.DISTOGRAM_BUCKETS)
-            )
+                Rearrange('b c h w -> b h w c')
+            ) if needs_upsample else nn.Identity(),
+            nn.Linear(dim, constants.DISTOGRAM_BUCKETS)
+        )
 
         # to coordinate output
 
@@ -497,6 +648,7 @@ class Alphafold2(nn.Module):
 
         with torch_default_dtype(torch.float64):
             self.structure_module_embeds = nn.Embedding(num_tokens, structure_module_dim)
+            self.atom_tokens_embed = nn.Embedding(len(ATOM_IDS), structure_module_dim)
 
             if use_se3_transformer:
                 self.structure_module = SE3TransformerWrapper(
@@ -544,19 +696,13 @@ class Alphafold2(nn.Module):
 
         # outer sum
 
-        x = rearrange(x, 'b i d -> b i () d') + rearrange(x, 'b j d-> b () j d') # create pair-wise residue embeds
-        x_mask = rearrange(mask, 'b i -> b i ()') + rearrange(mask, 'b j -> b () j') if exists(mask) else None
-
-        # flatten
-
-        seq_shape = x.shape
-        x = rearrange(x, 'b i j d -> b (i j) d')
-        x_mask = rearrange(x_mask, 'b i j -> b (i j)') if exists(mask) else None
+        x = rearrange(x, 'b i d -> b () i () d') + rearrange(x, 'b j d-> b () () j d') # create pair-wise residue embeds
+        x_mask = rearrange(mask, 'b i -> b () i ()') + rearrange(mask, 'b j -> b () () j') if exists(mask) else None
 
         # axial positional embedding
 
         pos_emb = rearrange(self.pos_emb(n_range), 'i d -> () i () d') + rearrange(self.pos_emb_ax(n_range), 'j d -> () () j d')
-        x += rearrange(pos_emb, 'b i j d -> b (i j) d')
+        x += pos_emb
 
         # embed multiple sequence alignment (msa)
 
@@ -578,7 +724,7 @@ class Alphafold2(nn.Module):
         if exists(msa_mask):
             msa_mask = rearrange(msa_mask, 'b m n -> b (m n)')
 
-        # trunk (template attention, wip)
+        # embed templates, if present
 
         if exists(templates_seq):
             assert exists(templates_coors), 'template residue coordinates must be supplied `templates_coors`'
@@ -597,24 +743,26 @@ class Alphafold2(nn.Module):
             # todo (make efficient)
 
             if exists(templates_sidechains):
-                shape = t_seq.shape
+                if self.use_se3_transformer:
+                    t_seq = self.template_sidechain_emb(
+                        t_seq,
+                        templates_sidechains,
+                        templates_coors,
+                        mask = templates_mask
+                    )
+                else:
+                    shape = t_seq.shape
+                    t_seq = rearrange(t_seq, 'b t n d -> (b t) n d')
+                    templates_coors = rearrange(templates_coors, 'b t n c -> (b t) n c')
+                    en_mask = rearrange(templates_mask, 'b t n -> (b t) n')
 
-                t_seq = rearrange(t_seq, 'b t n d-> (b t) n d ()')
+                    t_seq, _ = self.template_sidechain_emb(
+                        t_seq,
+                        templates_coors,
+                        mask = en_mask
+                    )
 
-                templates_sidechains = rearrange(templates_sidechains, 'b t n m -> (b t) n () m')
-                templates_sidechains = einsum('b n d m, d e -> b n e m', templates_sidechains, self.sidechains_proj)
-                templates_coors = rearrange(templates_coors, 'b t n m -> (b t) n m')
-
-                se3_templates_mask = rearrange(templates_mask, 'b t n -> (b t) n')
-
-                t_seq = self.template_sidechain_emb(
-                    {'0': t_seq, '1': templates_sidechains},
-                    templates_coors,
-                    mask = se3_templates_mask,
-                    return_type = 0
-                )
-
-                t_seq = t_seq.reshape(*shape)
+                    t_seq = t_seq.reshape(*shape)
 
             # embed template distances
 
@@ -623,50 +771,27 @@ class Alphafold2(nn.Module):
             t_seq = rearrange(t_seq, 'b t i d -> b t i () d') + rearrange(t_seq, 'b t j d -> b t () j d')
             t = t_seq + t_dist
 
-            template_shape = rearrange(t, 'b t h w d -> (b t) h w d').shape
-
             # template pos emb
 
             template_num_pos_emb = self.template_num_pos_emb(torch.arange(num_templates, device = device))
-
             t += rearrange(template_num_pos_emb, 't d-> () t () () d')
-            t = rearrange(t, 'b t h w d -> (b t) (h w) d')
 
-            pos_emb = rearrange(self.template_pos_emb(n_range), 'i d -> () i () d') + rearrange(self.template_pos_emb_ax(n_range), 'j d -> () () j d')
-            pos_emb = rearrange(pos_emb, 'b i j d -> b (i j) d')
+            pos_emb = rearrange(self.template_pos_emb(n_range), 'i d -> () () i () d') + rearrange(self.template_pos_emb_ax(n_range), 'j d -> () () () j d')
             t += pos_emb
 
-            if exists(templates_mask):
-                t_mask = rearrange(templates_mask, 'b t i -> b t i ()') * rearrange(templates_mask, 'b t j -> b t () j')
-                t_mask = rearrange(t_mask, 'b t h w -> (b t) (h w)')
-
             assert t.shape[-2:] == x.shape[-2:]
 
-            # template self attention followed by cross attention axially to primary sequence
+            x = torch.cat((x, t), dim = 1)
 
-            for (seq_self_attn, template_self_attn, cross_attn, ff) in self.template_attn_net:
-                x = seq_self_attn(x, mask = x_mask, shape = seq_shape)
-                t = template_self_attn(t, mask = t_mask, shape = template_shape) + t
+            if exists(templates_mask):
+                t_mask = rearrange(templates_mask, 'b t i -> b t i ()') * rearrange(templates_mask, 'b t j -> b t () j')
+                x_mask = torch.cat((x_mask, t_mask), dim = 1)
 
-                # concat template and primary sequence, for axial attention along the template dimension
-                # very similar to the attention across the time axis from https://arxiv.org/abs/2102.05095
+        # flatten
 
-                x = rearrange(x, 'b n d -> (b n) () d', n = n ** 2)
-                t = rearrange(t, '(b t) n d -> (b n) t d', t = num_templates)
-
-                y_mask = None
-                if exists(templates_mask) and exists(x_mask):
-                    t_mask_ = rearrange(t_mask, '(b t) n -> (b n) t', t = num_templates)
-                    mask_ = rearrange(x_mask, 'b n -> (b n) ()')
-                    y_mask = torch.cat((mask_, t_mask_), dim = 1)
-
-                y = torch.cat((x, t), dim = 1)
-                y = cross_attn(y, mask = y_mask) + y
-
-                x = rearrange(y[:, 0], '(b n) d -> b n d', n = n ** 2)
-                t = rearrange(y[:, 1:], '(b n) t d -> (b t) n d', n = (n ** 2))
-
-                t = ff(t) + t
+        seq_shape = x.shape
+        x = rearrange(x, 'b t i j d -> b (t i j) d')
+        x_mask = rearrange(x_mask, 'b t i j -> b (t i j)') if exists(mask) else None
 
         # trunk
 
@@ -679,9 +804,14 @@ class Alphafold2(nn.Module):
             msa_mask = msa_mask
         )
 
-        trunk_embeds = rearrange(x, 'b (h w) d -> b h w d', h = n)
+        # remove templates, if present
 
-        trunk_embeds = (trunk_embeds + rearrange(trunk_embeds, 'b i j d -> b j i d')) * 0.5  # symmetrize
+        x = x.view(seq_shape)
+        x = x[:, 0]
+
+        # embeds to distogram
+
+        trunk_embeds = (x + rearrange(x, 'b i j d -> b j i d')) * 0.5  # symmetrize
         distogram_logits = self.to_distogram_logits(trunk_embeds)
 
         if not self.predict_coords:
@@ -691,8 +821,8 @@ class Alphafold2(nn.Module):
 
         assert self.num_backbone_atoms > 1, 'must constitute to at least 3 atomic coordinates for backbone'
 
-        if self.num_backbone_atoms == 3:
-            N_mask, CA_mask, C_mask = scn_backbone_mask(seq, boolean = True)
+        if self.num_backbone_atoms >= 3:
+            N_mask, CA_mask, C_mask = scn_backbone_mask(seq, boolean = True, n_aa = self.num_backbone_atoms)
 
             cloud_mask = scn_cloud_mask(seq, boolean = True)
             flat_cloud_mask = rearrange(cloud_mask, 'b l c -> b (l c)')
@@ -713,17 +843,17 @@ class Alphafold2(nn.Module):
             C_mask = C_mask
         )
         coords = rearrange(coords_3d, 'b c n -> b n c')
+        # will init all sidechain coords to cbeta if present else c_alpha
         coords = sidechain_container(coords, n_aa = self.num_backbone_atoms, cloud_mask=cloud_mask)
-        coords = rearrange(coords[:, :, :self.num_backbone_atoms], 'b n l d -> b (n l) d')
-        # atom_tokens = scn_atom_embedd(seq)
+        coords = rearrange(coords, 'b n l d -> b (n l) d')
+        atom_tokens = scn_atom_embedd(seq) # not used for now, but could be
 
         
         trunk_embeds = self.trunk_to_structure_dim(trunk_embeds)
         x = reduce(trunk_embeds, 'b i j d -> b i d', 'mean')
         x += self.structure_module_embeds(seq)
-        x = repeat(x, 'b n d -> b n l d', l = self.num_backbone_atoms)
-
-        x += rearrange(self.backbone_pos_emb, 'l d -> () () l d')
+        x = repeat(x, 'b n d -> b n l d', l = cloud_mask.shape[-1])
+        x += self.atom_tokens_embed(atom_tokens)
         x = rearrange(x, 'b n l d -> b (n l) d')
 
         original_dtype = coords.dtype
@@ -731,9 +861,7 @@ class Alphafold2(nn.Module):
 
         with torch_default_dtype(torch.float64):
             for _ in range(self.structure_module_refinement_iters):
-                x, coords = self.structure_module(x, coords, mask = mask)
+                x, coords = self.structure_module(x, coords, mask = flat_chain_mask)
 
         coords.type(original_dtype)
-        return coords, mask
-
-
+        return coords
